/* --------------------------------------------------------------------------
 * File: mipex4.c
 * Version 12.4
 * --------------------------------------------------------------------------
 * Licensed Materials - Property of IBM
 * 5725-A06 5725-A29 5724-Y48 5724-Y49 5724-Y54 5724-Y55
 * Copyright IBM Corporation 2007, 2011. All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with
 * IBM Corp.
 * --------------------------------------------------------------------------
 */

/* mipex4.c - Reading in and optimizing a problem using a callback
              to log or interrupt or CPXsetterminate to interrupt */

/* To run this example, command line arguments are required.
   i.e.,   mipex4   filename option
                       where option is one of
                          t to use the time-limit-gap callback
                          l to use the logging callback
                          a to use the aborter

   Example:
       mipex4  mexample.mps l
*/

/* Bring in the CPLEX function declarations and the C library
   header file stdio.h with the following single include. */

#include <cplex.h>

/* Bring in the declarations for the string and character functions,
   math functions and malloc */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>

#define EPSZERO        1.0E-10
#define BUFSIZE 16

/* Define the structures used to communicate info to the callbacks */

struct timeliminfo {
    double timestart;
    double timelim;
    double acceptablegap;
    int    aborted;
};
typedef struct timeliminfo TIMELIMINFO, *TIMELIMINFOptr;

struct loginfo {
   double lastincumbent;
   int    lastlog;
   int    numcols;
};
typedef struct loginfo LOGINFO, *LOGINFOptr;


static int num_SNPs ;
static CPXENVptr env = NULL;
static CPXLPptr  lp = NULL;
static int surplus, numcols;
static int solstat ;
static char **cur_colname = NULL;
static char *cur_colnamestore = NULL;
static int cur_colnamespace;
static char tempName[100], *s_x, *s_y ;

static int is_4_choice, isHard ;

/* Include declarations for functions in this program */

static void
   free_and_null (char **ptr),
   usage (char *progname);

static int CPXPUBLIC
   timelimcallback (CPXCENVptr env, void *cbdata, int wherefrom,
                    void *cbhandle);

static int CPXPUBLIC
   logcallback (CPXCENVptr env, void *cbdata, int wherefrom,
                void *cbhandle);

static char* hapFile ;

static int *sameCol = NULL ;
static int *diffCol = NULL ;

int main (int argc, char *argv[]) {
   FILE *fp ;
   int     uselogcallback = 0;
   LOGINFO myloginfo;

   int         usetimelimcallback = 0;
   TIMELIMINFO mytimeliminfo;

   int          useterminate = 0;
   volatile int terminator;

   int       status = 0, h, j, i ;
   double   *x     = NULL, objval, *incx = NULL ;
   char *solFile, *sameColFile, *diffColFile, *t_x, *t_y ;
   struct stat st ;
   long timeLimit ;
   int isAll, log_call_back ; 

   /* Check the command line arguments */

   if ( argc != 7 && argc != 8 ) {
      usage (argv[0]);
      exit(1) ;
   }

	num_SNPs = atoi( argv[2] ) ;
	is_4_choice = atoi( argv[3] ) ;
	isHard = atoi ( argv[4] ) ;
	isAll = atoi ( argv[5] ) ;
	log_call_back = atoi ( argv[6] ) ;

	j = strlen( argv[1] ) ;
	solFile = (char *) malloc( (j + 10) * sizeof(char) ) ;
	hapFile = (char *) malloc( (j + 10) * sizeof(char) ) ;
	sameColFile = (char *) malloc( (j + 10) * sizeof(char) ) ;
	diffColFile = (char *) malloc( (j + 10) * sizeof(char) ) ;
	s_x = (char *) malloc( (num_SNPs + 1) * sizeof(char) ) ;
	s_y = (char *) malloc( (num_SNPs + 1) * sizeof(char) ) ;
	t_x = (char *) malloc( (num_SNPs + 1) * sizeof(char) ) ;
	t_y = (char *) malloc( (num_SNPs + 1) * sizeof(char) ) ;
	if (solFile == NULL || hapFile == NULL || sameColFile == NULL 
			|| diffColFile == NULL || s_x == NULL || s_y == NULL 
			|| t_x == NULL || t_y == NULL) { 
		puts("Out of memory.") ;
		exit(1) ;
	}
	strcpy(solFile, argv[1]) ;
	strcpy(&solFile[j - 2], "sol") ;
	strcpy(hapFile, argv[1]) ;
	strcpy(&hapFile[j - 2], "hap") ;
	strcpy(sameColFile, argv[1]) ;
	strcpy(&sameColFile[j - 2], "sam") ;
	strcpy(diffColFile, argv[1]) ;
	strcpy(&diffColFile[j - 2], "dif") ;
	puts(solFile) ;
	puts(hapFile) ;
	puts(sameColFile) ;
	puts(diffColFile) ;

    if ( stat(sameColFile, &st) == 0 ) { 
	   fp = fopen(sameColFile, "r") ;
	   if (fp == NULL) { 
    	   printf("failed to open the output file.") ;
	       return 1 ;
	   }
		sameCol = (int *)malloc( sizeof(int) * num_SNPs ) ;
		if (sameCol == NULL) {
			puts("out of memory.") ;
			fclose(fp) ;
			return 1 ;
		}
		for ( j = 0 ; j < num_SNPs ; j++) 
			sameCol[j] = 0 ;
		while ( fscanf(fp, "%d%d", &h, &j ) != EOF ) {
			sameCol[j - 1] = h ;
		}
		fclose(fp) ;
	}

    if ( stat(diffColFile, &st) == 0 ) { 
	   fp = fopen(diffColFile, "r") ;
	   if (fp == NULL) { 
    	   printf("failed to open the output file.") ;
	       return 1 ;
	   }
		diffCol = (int *)malloc( sizeof(int) * num_SNPs ) ;
		if (diffCol == NULL) {
			puts("out of memory.") ;
			fclose(fp) ;
			return 1 ;
		}
		for ( j = 0 ; j < num_SNPs ; j++) 
			diffCol[j] = 0 ;
		while ( fscanf(fp, "%d%d", &h, &j ) != EOF ) {
			diffCol[j - 1] = h ;
		}
		fclose(fp) ;
	}

	if ( log_call_back == 1 ) 
	   uselogcallback = 1;
	else
	   uselogcallback = 0;

   /* Initialize the CPLEX environment */

   env = CPXopenCPLEX (&status);

   /* If an error occurs, the status value indicates the reason for
      failure.  A call to CPXgeterrorstring will produce the text of
      the error message.  Note that CPXopenCPLEX produces no output,
      so the only way to see the cause of the error is to use
      CPXgeterrorstring.  For other CPLEX routines, the errors will
      be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON.  */

   if ( env == NULL ) {
      char  errmsg[CPXMESSAGEBUFSIZE];
      fprintf (stderr, "Could not open CPLEX environment.\n");
      CPXgeterrorstring (env, status, errmsg);
      fprintf (stderr, "%s", errmsg);
      goto TERMINATE;
   }

   /* set parameters */

   status = CPXsetdblparam (env, CPX_PARAM_EPGAP, 0.0);
   if ( status ) {
      fprintf (stderr,
               "Failure to set the EP gap, error %d.\n", status);
      goto TERMINATE;
   }

   status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
   if ( status ) {
      fprintf (stderr,
               "Failure to turn on screen indicator, error %d.\n", status);
      goto TERMINATE;
   }

   status = CPXsetintparam (env, CPX_PARAM_ADVIND, 2);
   if ( status ) {
      fprintf (stderr,
               "Failure to set the MIP start switch, error %d.\n", status);
      goto TERMINATE;
   }

/*
   status = CPXsetintparam (env, CPX_PARAM_THREADS, NUM_THREADS);
   if ( status ) {
      fprintf (stderr, 
               "Failure to set the number of threads, error %d.\n", status);
      goto TERMINATE;
   }
*/

   //if (isHard == 0) {
	   status = CPXsetintparam (env, CPX_PARAM_MEMORYEMPHASIS, CPX_ON);
	   if ( status ) {
	      fprintf (stderr, 
               "Failure to turn on memory emphasis, error %d.\n", status);
    	  goto TERMINATE;
	   }
   //}
   if (isHard == 1) {
	   status = CPXsetintparam (env, CPX_PARAM_NODEFILEIND, 3);
	   if ( status ) {
	      fprintf (stderr, 
               "Failure to turn on node file indicator, error %d.\n", status);
    	  goto TERMINATE;
	   }
	   status = CPXsetintparam (env, CPX_PARAM_VARSEL, CPX_VARSEL_STRONG);
   	   if ( status ) {
	      fprintf (stderr, 
               "Failure to turn on variable selection, error %d.\n", status);
    	  goto TERMINATE;
	   }
       status = CPXsetintparam (env, CPX_PARAM_MIPEMPHASIS, CPX_MIPEMPHASIS_BESTBOUND);
	   if ( status ) {
    	  fprintf (stderr, 
               "Failure to turn on optimality emphasis, error %d.\n", status);
	      goto TERMINATE;
   	  }
   }

	if (argc == 8) {
		timeLimit = atol( argv[7] ) ;
		if ( timeLimit > 0 ) {
			status = CPXsetdblparam(env, CPX_PARAM_TILIM, timeLimit) ; 
			if (status) { 
		       fprintf (stderr, "Failure to set time limit, error %d.\n", status);
	    	   goto TERMINATE;
			}
		}
	}

	if (isAll) {
	   status = CPXsetdblparam (env, CPX_PARAM_SOLNPOOLAGAP, 0.0);
	   if ( status ) {
    	  fprintf (stderr,
               "Failure to set SOLNPOOLAGAP, error %d.\n", status);
	      goto TERMINATE;
	   }
	   status = CPXsetintparam (env, CPX_PARAM_SOLNPOOLINTENSITY, 4);
	   if ( status ) {
    	  fprintf (stderr,
               "Failure to set SOLNPOOLINTENSITY, error %d.\n", status);
	      goto TERMINATE;
	   }
	   status = CPXsetintparam (env, CPX_PARAM_POPULATELIM, 21000000L);
	   if ( status ) {
    	  fprintf (stderr,
               "Failure to set POPULATELIM, error %d.\n", status);
	      goto TERMINATE;
	   }
    }


   /* Create the problem, using the filename as the problem name */

   lp = CPXcreateprob (env, &status, argv[1]);

   /* A returned pointer of NULL may mean that not enough memory
      was available or there was some other problem.  In the case of
      failure, an error message will have been written to the error
      channel from inside CPLEX.  In this example, the setting of
      the parameter CPX_PARAM_SCRIND causes the error message to
      appear on stdout.  Note that most CPLEX routines return
      an error code to indicate the reason for failure.   */

   if ( lp == NULL ) {
      fprintf (stderr, "Failed to create LP.\n");
      goto TERMINATE;
   }

   /* Now read the file, and copy the data into the created lp */

   status = CPXreadcopyprob (env, lp, argv[1], NULL);
   if ( status ) {
      fprintf (stderr, "Failed to read and copy the problem data.\n");
      goto TERMINATE;
   }

   if ( stat(solFile, &st) == 0 ) { 
	  status = CPXreadcopymipstarts(env, lp, solFile) ;
	  if ( status ) {
    	  fprintf (stderr, "Failed to read the starting solution.\n");
          goto TERMINATE;
	  }
   }

   /* Set overall node limit in case callback conditions are not met */
//   status = CPXsetintparam (env, CPX_PARAM_NODELIM, 5000);
//   if ( status ) goto TERMINATE;

   if ( usetimelimcallback ) {
      double t;
      status = CPXgettime (env, &t);
      if ( status ) {
         fprintf (stderr, "Failed to initialize timer.\n");
         goto TERMINATE;
      }
      mytimeliminfo.acceptablegap = 10.0;
      mytimeliminfo.aborted       = 0;
      mytimeliminfo.timestart     = t;
      mytimeliminfo.timelim       = 1.0;

      status = CPXsetinfocallbackfunc (env, timelimcallback, &mytimeliminfo);
      if ( status ) {
         fprintf (stderr, "Failed to set time limit callback function.\n");
         goto TERMINATE;
      }
   }
   else if ( uselogcallback ) {
      myloginfo.numcols       = CPXgetnumcols (env, lp);
      myloginfo.lastincumbent = CPXgetobjsen (env, lp) * 1e+35;
      myloginfo.lastlog       = -10000;
      status = CPXsetinfocallbackfunc (env, logcallback, &myloginfo);
      if ( status ) {
         fprintf (stderr, "Failed to set logging callback function.\n");
         goto TERMINATE;
      }
      /* Turn off CPLEX logging */
//      status = CPXsetintparam (env, CPX_PARAM_MIPDISPLAY, 0);
//      if ( status )  goto TERMINATE;
   }
   else if ( useterminate) {
      status = CPXsetterminate (env, &terminator);
      if ( status ) {
         fprintf (stderr, "Failed to set terminator.\n");
         goto TERMINATE;
      }
      /* Typically, you would pass the terminator variable to
         another thread or pass it to an interrupt handler,
         and  monitor for some event to occur.  When it does,
         set terminator to a non-zero value.
 
         To illustrate its use without creating a thread or
         an interrupt handler, terminate immediately by setting
         terminator before the solve.
      */
      terminator = 1;
   }

	  /* Now get the column names for the problem.  First we determine how
	     much space is used to hold the names, and then do the allocation.
	     Then we call CPXgetcolname() to get the actual names. */ 

	  numcols = CPXgetnumcols (env, lp);
	  status = CPXgetcolname (env, lp, NULL, NULL, 0, &surplus, 0, numcols-1);
	  if (( status != CPXERR_NEGATIVE_SURPLUS ) &&
	      ( status != 0 )                         )  {
	     fprintf (stderr, 
	              "Could not determine amount of space for column names.\n");
	     goto TERMINATE;
	  }

	  cur_colnamespace = - surplus;
	  if ( cur_colnamespace > 0 ) {
	     cur_colname      = (char **) malloc (sizeof(char *) * numcols);
	     cur_colnamestore = (char *)  malloc (cur_colnamespace);
	     if ( cur_colname      == NULL ||
	          cur_colnamestore == NULL   ) {
	        fprintf (stderr, "Failed to get memory for column names.\n");
		    status = -1;
	        goto TERMINATE;
	     }
    	 status = CPXgetcolname (env, lp, cur_colname, cur_colnamestore, 
	                             cur_colnamespace, &surplus, 0, numcols-1);
	     if ( status ) {
	        fprintf (stderr, "CPXgetcolname failed.\n");
	        goto TERMINATE;
	     }
	  }
	  else {
	     printf ("No names associated with problem.  Using Fake names.\n");
	  }

   /* Optimize the problem and obtain solution. */

   if (isAll == 0)
	   status = CPXmipopt (env, lp);
   else
	   status = CPXpopulate (env, lp);

   if ( status ) {
      fprintf (stderr, "Failed to optimize MIP.\n");
      goto TERMINATE;
   }

   solstat = CPXgetstat (env, lp);
   printf ("Solution status %d.\n", solstat);

	if (solstat != CPXMIP_OPTIMAL && argc == 8) {
		printf("***** CAUTION: Failed to solve %s optimally in %ld seconds *****\n", 
				argv[1], timeLimit) ;
	} 

   status = CPXgetobjval (env, lp, &objval);
   if ( status ) {
      fprintf (stderr,"No MIP objective value available.  Exiting...\n");
      goto TERMINATE;
   }

   printf ("Solution value  = %f\n\n", objval);

   /* Allocate space for solution */
   x = (double *) malloc (numcols*sizeof(double));
   if ( x == NULL ) {
      fprintf (stderr, "No memory for solution values.\n");
      goto TERMINATE;
   }

   status = CPXgetx (env, lp, x, 0, numcols-1);
   if ( status ) {
      fprintf (stderr, "Failed to obtain solution.\n");
      goto TERMINATE;
   }

   /* Write out the solution */
// puts("y:") ;
   for (h = 1; h <= num_SNPs; h++) {
	   for (j = 0 ; j < 10 ; j++) 
			tempName[j] = '\0' ;
	   tempName[0] = 'y' ;
	   sprintf(&tempName[1], "%d", h); 
	   for (j = 0; j < numcols; j++) {
    	  if ( strcmp(cur_colname[j], tempName) == 0 ) {
//        	 printf ("%-16s:  ", cur_colname[j]);
			 if (x[j] > 0.5) s_y[h - 1] = '1' ;
			 else s_y[h - 1] = '0' ;
			 break ;
		  }
       }
	   if ( j == numcols ) 
			s_y[h - 1] = '0' ;
   }

/*
   puts("all:") ;
   for (j = 0; j < numcols; j++) {
      if (strlen(cur_colname[j]) == 0 || cur_colname[j][0] == 't' || cur_colname[j][0] == 'z')
		 continue ;
      if ( cur_colnamespace > 0 ) {
         printf ("%-16s:  ", cur_colname[j]);
      }
      else {
         printf ("Fake%-6.6d      :  ", j);;
      }
      printf ("%17.10g\n", x[j]);
   }
   puts("") ;
*/

   if (is_4_choice == 0) goto OUTPUT_y;

// puts("x:") ;
   for (h = 1; h <= num_SNPs; h++) {
	   for (j = 0 ; j < 10 ; j++) 
			tempName[j] = '\0' ;
	   tempName[0] = 'x' ;
	   sprintf(&tempName[1], "%d", h); 
	   for (j = 0; j < numcols; j++) {
    	  if ( strcmp(cur_colname[j], tempName) == 0 ) {
//        	 printf ("%-16s:  ", cur_colname[j]);
			 if (x[j] > 0.5) s_x[h - 1] = '1' ;
			 else s_x[h - 1] = '0' ;
			 break ;
		  }
       }
	   if ( j == numcols ) 
			s_x[h - 1] = '0' + '1' - s_y[h - 1] ;
   }

OUTPUT_y:
   fp = fopen(hapFile, "w") ;
   if (fp == NULL) { 
       printf("failed to open the output file.") ;
       return 1 ;
   }

   if (sameCol != NULL) {
	   for (j = 0 ; j < num_SNPs ; j++) {
		   if ( sameCol[j] > 0 ) {
			   s_y[j] = s_y[sameCol[j] - 1] ;
		   } 
	   }
   }
   if (diffCol != NULL) {
	   for (j = 0 ; j < num_SNPs ; j++) {
		   if ( diffCol[j] > 0 ) {
			   s_y[j] = '0' + '1' - s_y[diffCol[j] - 1] ;
		   } 
	   }
   }

   s_y[num_SNPs] = '\0' ;
   fprintf(fp, "%s\n", s_y) ;
   if (is_4_choice) {
	   if (sameCol != NULL) {
		   for (j = 0 ; j < num_SNPs ; j++) {
			   if ( sameCol[j] > 0 ) {
				   s_x[j] = s_x[sameCol[j] - 1] ;
			   }
		   }
	   }
	   if (diffCol != NULL) {
		   for (j = 0 ; j < num_SNPs ; j++) {
			   if ( diffCol[j] > 0 ) {
				   s_x[j] = '0' + '1' - s_x[diffCol[j] - 1] ;
			   } 
		   }
	   }
	   s_x[num_SNPs] = '\0' ;
   	   fprintf(fp, "%s\n", s_x) ;
   }
   fclose(fp) ;

	if ( isAll) { /* Get the number of solutions in the solution pool */
	   int numsol, numsolreplaced, numdiff ;
	   double meanobjval ;

	   numsol = CPXgetsolnpoolnumsolns (env, lp);   
	   printf ("The solution pool contains %d solutions.\n", numsol);

	   /* Some solutions are deleted from the pool because of the solution
    	  pool relative gap parameter */

	   numsolreplaced = CPXgetsolnpoolnumreplaced (env, lp);
	   printf ("%d solutions were removed due to the solution pool relative gap parameter.\n",
    	      numsolreplaced);

	   printf ("In total, %d solutions were generated.\n",
           numsol + numsolreplaced);
   
	   /* Get the average objective value of solutions in the solution pool */
	   status = CPXgetsolnpoolmeanobjval (env, lp, &meanobjval);
	   printf ("The average objective value of the solutions is %.10g.\n\n",
    	      meanobjval);

	   /* Write out the objective value of each solution and its
    	  difference to the incumbent */

	   incx = (double *) malloc (numcols*sizeof(double));
	   if ( incx == NULL ) {
    	  fprintf (stderr, "No memory for solution values.\n");
	      goto TERMINATE;
	   }

	   printf ("Solution        Objective   Number of variables\n");
	   printf ("                value       that differ compared to\n");
	   printf ("                            the incumbent\n");

	   for (i = 0; i < numsol; i++) {
    	  char namei[BUFSIZE];
	      int  surplus;

    	  /* Write out objective value */

	      CPXgetsolnpoolsolnname (env, lp, namei, BUFSIZE, &surplus, i);
    	  printf ("%-15s ", namei); 


	      status = CPXgetsolnpoolobjval (env, lp, i, &objval);
    	  if ( status ) {
	         fprintf (stderr,
        	          "Failed to obtain objective value for solution %d.\n", i);
    	     goto TERMINATE;
	      }
    	  printf ("%.10g         ", objval);

	      status = CPXgetsolnpoolx (env, lp, i, incx, 0, numcols-1);
    	  if ( status ) {
	         fprintf (stderr, "Failed to obtain solution %d.\n", i);
        	 goto TERMINATE;
    	  }

		  for (h = 1; h <= num_SNPs; h++) {
	   		for (j = 0 ; j < 10 ; j++) 
				tempName[j] = '\0' ;
	   		tempName[0] = 'y' ;
		    sprintf(&tempName[1], "%d", h); 
		    for (j = 0; j < numcols; j++) {
    	 		if ( strcmp(cur_colname[j], tempName) == 0 ) {
					if (incx[j] > 0.5) t_y[h - 1] = '1' ;
					else t_y[h - 1] = '0' ;
					break ;
				}
       		}
		    if ( j == numcols ) 
				t_y[h - 1] = '0' ;
   		  }

		  if ( is_4_choice) {
			  for (h = 1; h <= num_SNPs; h++) {
	   			for (j = 0 ; j < 10 ; j++) 
					tempName[j] = '\0' ;
		   		tempName[0] = 'x' ;
			    sprintf(&tempName[1], "%d", h); 
		    	for (j = 0; j < numcols; j++) {
    		 		if ( strcmp(cur_colname[j], tempName) == 0 ) {
						if (incx[j] > 0.5) t_x[h - 1] = '1' ;
						else t_x[h - 1] = '0' ;
						break ;
					}
       			}
			    if ( j == numcols ) 
					t_x[h - 1] = '0' + '1' - t_y[h - 1] ;
	   		  }
		   } else {
			  for (h = 0 ; h < num_SNPs ; h++) 
				t_x[h] = '0' + '1' - t_y[h] ;
		   }

	       if (sameCol != NULL) 
		       for (j = 0 ; j < num_SNPs ; j++) 
			       if ( sameCol[j] > 0 ) {
				       t_y[j] = t_y[sameCol[j] - 1] ;
				       t_x[j] = t_x[sameCol[j] - 1] ;
				   }
		   if (diffCol != NULL) 
	   		   for (j = 0 ; j < num_SNPs ; j++) 
				   if ( diffCol[j] > 0 ) {
					   t_y[j] = '0' + '1' - t_y[diffCol[j] - 1] ;
					   t_x[j] = '0' + '1' - t_x[diffCol[j] - 1] ;
				   }

	      /* Compute the number of variables that differ in the solution
        	 and in the incumbent */
    	  numdiff = 0;
		  for (j = 0 ; j < num_SNPs ; j++) {
			 if ( s_x[j] != t_x[j] ) numdiff++ ;
			 if ( s_y[j] != t_y[j] ) numdiff++ ;
		  }
    	  h = 0;
		  for (j = 0 ; j < num_SNPs ; j++) {
			 if ( s_x[j] != t_y[j] ) h++ ;
			 if ( s_y[j] != t_x[j] ) h++ ;
		  }
		  if (numdiff > h) numdiff = h ;
		  printf ("%d / %d\n", numdiff, 2 * num_SNPs);
	   }
	}

TERMINATE:

   /* Free up the problem as allocated by CPXcreateprob, if necessary */

   if ( lp != NULL ) {
      int xstatus = CPXfreeprob (env, &lp);
      if ( xstatus ) {
         fprintf (stderr, "CPXfreeprob failed, error code %d.\n", xstatus);
         status = xstatus;
      }
   }

   /* Free up the CPLEX environment, if necessary */

   if ( env != NULL ) {
      int xstatus = CPXcloseCPLEX (&env);

      /* Note that CPXcloseCPLEX produces no output,
         so the only way to see the cause of the error is to use
         CPXgeterrorstring.  For other CPLEX routines, the errors will
         be seen if the CPX_PARAM_SCRIND indicator is set to CPX_ON. */

      if ( status ) {
         char  errmsg[CPXMESSAGEBUFSIZE];
         fprintf (stderr, "Could not close CPLEX environment.\n");
         CPXgeterrorstring (env, status, errmsg);
         fprintf (stderr, "%s", errmsg);
         status = xstatus;
      }
   }

	free(solFile) ;
	free(hapFile) ;
	free(sameColFile) ;
	free(diffColFile) ;
	free(s_x) ;
	free(s_y) ;
	free(t_x) ;
	free(t_y) ;
	free(sameCol) ;
    free_and_null ((char **) &incx);
    free_and_null ((char **) &x);
   return (status);

}  /* END main */


/* Spend at least timelim seconds on optimization, but once
   this limit is reached, quit as soon as the solution is acceptable */

static int CPXPUBLIC
timelimcallback (CPXCENVptr env, void *cbdata, int wherefrom, void *cbhandle)
{
   int status = 0;

   TIMELIMINFOptr info = (TIMELIMINFOptr) cbhandle;
   int            hasincumbent = 0;

   status = CPXgetcallbackinfo (env, cbdata, wherefrom,
                                CPX_CALLBACK_INFO_MIP_FEAS, &hasincumbent);
   if ( status )  goto TERMINATE;

   if ( !info->aborted  &&  hasincumbent ) {
      double gap;
      double timenow;

      status = CPXgetcallbackinfo (env, cbdata, wherefrom,
                                   CPX_CALLBACK_INFO_MIP_REL_GAP, &gap);
      if ( status )  goto TERMINATE;

      /* Turn the gap into a percentage */
      gap *= 100.0;

      status = CPXgettime (env, &timenow);
      if ( status )  goto TERMINATE;

      if ( timenow - info->timestart > info->timelim  &&
           gap < info->acceptablegap                    ) {
         fprintf (stderr,
     "Good enough solution at time %.2fsec, gap = %g%%, quitting\n",
                  timenow - info->timestart, gap);

         /* callback may be called again during the clean up phase after the
            abort has been issued, so remember that abort processing has
            already occurred. */

         info->aborted = 1;

         /* nonzero status causes abort */

         status = 1;
      }
   }

TERMINATE:

   return (status);

} /* END timelimcallback */


/* Log new incumbents if they are at better than the old by a
   relative tolerance of 1e-5; also log progress info every
   100 nodes. */

static int CPXPUBLIC
logcallback (CPXCENVptr env, void *cbdata, int wherefrom, void *cbhandle)
{
   int status = 0;

   LOGINFOptr info = (LOGINFOptr) cbhandle;
   int        hasincumbent = 0;
   int        newincumbent = 0;
   int        nodecnt;
   int        nodesleft;
   double     objval;
   double     bound;
   double     *x = NULL;

   status = CPXgetcallbackinfo (env, cbdata, wherefrom,
                                CPX_CALLBACK_INFO_NODE_COUNT, &nodecnt);
   if ( status )  goto TERMINATE;

   status = CPXgetcallbackinfo (env, cbdata, wherefrom,
                                CPX_CALLBACK_INFO_NODES_LEFT, &nodesleft);
   if ( status )  goto TERMINATE;

   status = CPXgetcallbackinfo (env, cbdata, wherefrom,
                                CPX_CALLBACK_INFO_MIP_FEAS, &hasincumbent);
   if ( status )  goto TERMINATE;

   if ( hasincumbent ) {
      status = CPXgetcallbackinfo (env, cbdata, wherefrom,
                                   CPX_CALLBACK_INFO_BEST_INTEGER, &objval);
      if ( status )  goto TERMINATE;

      if ( fabs(info->lastincumbent - objval) > 1e-5*(1.0 + fabs(objval)) ) {
         newincumbent = 1;
         info->lastincumbent = objval;
      }
   }

   if ( nodecnt >= info->lastlog + 100  ||  newincumbent ) {
      status = CPXgetcallbackinfo (env, cbdata, wherefrom,
                                   CPX_CALLBACK_INFO_BEST_REMAINING, &bound);
      if ( status )  goto TERMINATE;

      if ( !newincumbent )  info->lastlog = nodecnt;

      printf ("Nodes = %d(%d)  Best objective = %g", nodecnt, nodesleft, bound);
      if ( hasincumbent )  printf ("  Incumbent objective = %g\n", objval);
      else                 printf ("\n");
   }

   if ( newincumbent ) {
      int j, h ;
	  FILE *fp ;
//      int numcols = info->numcols;
      x = (double *) malloc (numcols*sizeof(double));
      if ( x == NULL ) {
         status = CPXERR_NO_MEMORY;
         goto TERMINATE;
      }
      status = CPXgetcallbackincumbent (env, cbdata, wherefrom,
                                        x, 0, numcols-1);
      if ( status )  goto TERMINATE;

	  /* Write out the solution */
      printf ("Found a new incumbent solution.\n");

	  for (h = 1; h <= num_SNPs; h++) {
	   for (j = 0 ; j < 100 ; j++) 
			tempName[j] = '\0' ;
	   tempName[0] = 'y' ;
	   sprintf(&tempName[1], "%d", h); 
	   for (j = 0; j < numcols; j++) {
    	  if ( strcmp(cur_colname[j], tempName) == 0 ) {
//        	 printf ("%-16s:  ", cur_colname[j]);
			 if (x[j] > 0.5) s_y[h - 1] = '1';
			 else s_y[h - 1] = '0' ;
			 break ;
		  }
       }
	   if ( j == numcols ) 
			s_y[h - 1] = '0' ;
	  }

	  if (is_4_choice == 0) goto SAVE_y ;

	  for (h = 1; h <= num_SNPs; h++) {
	   for (j = 0 ; j < 100 ; j++) 
			tempName[j] = '\0' ;
	   tempName[0] = 'x' ;
	   sprintf(&tempName[1], "%d", h); 
	   for (j = 0; j < numcols; j++) {
    	  if ( strcmp(cur_colname[j], tempName) == 0 ) {
//        	 printf ("%-16s:  ", cur_colname[j]);
			 if (x[j] > 0.5) s_x[h - 1] = '1' ;
			 else s_x[h - 1] = '0' ;
			 break ;
		  }
       }
	   if ( j == numcols ) 
			s_x[h - 1] = '0' + '1' - s_y[h - 1] ;
	  }

SAVE_y: 
	  fp = fopen(hapFile, "w") ;
	  if (fp == NULL) { 
	      printf("failed to open the output file.") ;
       return 1 ;
	  }
	  if (is_4_choice) {
	   	  if (sameCol != NULL) {
			  for (j = 0 ; j < num_SNPs ; j++) {
			   	  if ( sameCol[j] > 0 ) {
			    	  s_x[j] = s_x[sameCol[j] - 1] ;
				  }
		 	 }
	  	  }
		   if (diffCol != NULL) {
			   for (j = 0 ; j < num_SNPs ; j++) {
				   if ( diffCol[j] > 0 ) {
					   s_x[j] = '0' + '1' - s_x[diffCol[j] - 1] ;
				   } 
			   }
		   }
		  s_x[num_SNPs] = '\0' ;
		  fprintf(fp, "%s\n", s_x) ;
	  }

   	  if (sameCol != NULL) {
		  for (j = 0 ; j < num_SNPs ; j++) {
		   	  if ( sameCol[j] > 0 ) {
			      s_y[j] = s_y[sameCol[j] - 1] ;
			  }
		  }
	  }
	   if (diffCol != NULL) {
		   for (j = 0 ; j < num_SNPs ; j++) {
			   if ( diffCol[j] > 0 ) {
				   s_y[j] = '0' + '1' - s_y[diffCol[j] - 1] ;
			   } 
		   }
	   }
	  s_y[num_SNPs] = '\0' ;
	  fprintf(fp, "%s\n", s_y) ;
	  fclose(fp) ;

/*      for (j = 0; j < numcols; j++) {
         if ( fabs(x[j]) > 1e-6 ) {
            printf ("  Column %d:  %g\n", j, x[j]);
         }
      }*/
   }

TERMINATE:

   free_and_null ((char **) &x);
   return (status);

} /* END logcallback */


/* This simple routine frees up the pointer *ptr, and sets *ptr to NULL */

static void
free_and_null (char **ptr)
{
   if ( *ptr != NULL ) {
      free (*ptr);
      *ptr = NULL;
   }
} /* END free_and_null */


static void
usage (char *progname)
{
   fprintf (stderr, "Usage: %s filename num_SNPs is_4_choice isHard isAll log_call_back [time_limit]\n", progname);
   fprintf (stderr, "   where filename is a file with extension \n");
   fprintf (stderr, "      \t MPS, SAV, or LP (lower case is allowed)\n");
   fprintf (stderr, "      \t num_SNPs is the number of SNPs,\n");
   fprintf (stderr, "      \t is_4_choice is either 0 (2-choice)  or 1 (4-choice),\n");
   fprintf (stderr, "      \t isHard is either 0 (easy MIP)  or 1 (hard MIP),\n");
   fprintf (stderr, "      \t isAll is either 0 (one solution)  or 1 (all solutions), \n");
   fprintf (stderr, "      \t log_call_back is either 1 (log intermediate solution)  or 1 (otherwise), and\n");
   fprintf (stderr, "      \t time_limit is a limit (in seconds) on the running time.\n");
   fprintf (stderr, " Exiting...\n");
} /* END usage */
